@@ -66,8 +66,8 @@ QMap<IProject*, QVector<QUrl>> Helper::cachedCustomIncludes;
 QMap<IProject*, QVector<QUrl>> Helper::cachedSearchPaths;
 QVector<QUrl> Helper::projectSearchPaths;
 QStringList Helper::dataDirs;
-QString Helper::documentationFile;
-DUChainPointer<TopDUContext> Helper::documentationFileContext = DUChainPointer<TopDUContext>(0);
+IndexedString Helper::documentationFile;
+DUChainPointer<TopDUContext> Helper::documentationFileContext = DUChainPointer<TopDUContext>(nullptr);
 QStringList Helper::correctionFileDirs;
 QString Helper::localCorrectionFileDir;
 QMutex Helper::cacheMutex;
@@ -88,7 +88,7 @@ void Helper::scheduleDependency(const IndexedString& dependency, int betterThanP
     }
     if ( needsReschedule ) {
         bgparser->addDocument(dependency, TopDUContext::ForceUpdate, betterThanPriority - 1,
-                              0, ParseJob::FullSequentialProcessing);
+                              nullptr, ParseJob::FullSequentialProcessing);
     }
 }
 
@@ -115,7 +115,7 @@ Declaration* Helper::accessAttribute(const AbstractType::Ptr accessed,
                                      const TopDUContext* topContext)
 {
     if ( ! accessed ) {
-        return 0;
+        return nullptr;
     }
     // if the type is unsure, search all the possibilities (but return the first match)
     auto structureTypes = Helper::filterType<StructureType>(accessed,
@@ -176,46 +176,72 @@ Helper::FuncInfo Helper::functionForCalled(Declaration* called, bool isAlias)
     return { dynamic_cast<FunctionDeclaration*>(attr), isAlias };
 }
 
-Declaration* Helper::declarationForName(const QualifiedIdentifier& identifier, const RangeInRevision& nodeRange,
+Declaration* Helper::declarationForName(const QString& name, const CursorInRevision& location,
                                         KDevelop::DUChainPointer<const DUContext> context)
 {
-    QList<Declaration*> declarations;
-    QList<Declaration*> localDeclarations;
-    QList<Declaration*> importedLocalDeclarations;
-    {
-        DUChainReadLocker lock(DUChain::lock());
-        if ( context.data() == context->topContext() && nodeRange.isValid() ) {
-            declarations = context->topContext()->findDeclarations(identifier, nodeRange.end);
-        }
-        else {
-            declarations = context->topContext()->findDeclarations(identifier, CursorInRevision::invalid());
-        }
-        localDeclarations = context->findLocalDeclarations(identifier.last(), nodeRange.end, 0,
-                                                           AbstractType::Ptr(0), DUContext::DontResolveAliases);
-        importedLocalDeclarations = context->findDeclarations(identifier.last(), nodeRange.end);
-    }
-    Declaration* declaration = 0;
-    if ( localDeclarations.length() ) {
-        declaration = localDeclarations.last();
-    }
-    else if ( importedLocalDeclarations.length() ) {
-        // don't use declarations from class decls, they must be referenced through "self.<foo>"
-        do {
-            declaration = importedLocalDeclarations.last();
-            importedLocalDeclarations.pop_back();
-            if ( !declaration || (declaration->context()->type() == DUContext::Class && context->type() != DUContext::Function) ) {
-                declaration = 0;
-            }
-            if ( importedLocalDeclarations.isEmpty() ) {
-                break;
-            }
-        } while ( ! importedLocalDeclarations.isEmpty() );
+    DUChainReadLocker lock(DUChain::lock());
+    auto identifier = KDevelop::Identifier(name);
+    auto localDeclarations = context->findLocalDeclarations(identifier, location, nullptr,
+                                                            AbstractType::Ptr(), DUContext::DontResolveAliases);
+    if ( !localDeclarations.isEmpty() ) {
+        return localDeclarations.last();
     }
 
-    if ( !declaration && declarations.length() ) {
-        declaration = declarations.last();
+    QList<Declaration*> declarations;
+    const DUContext* currentContext = context.data();
+    bool findInNext = true, findBeyondUse = false;
+    do {
+        if (findInNext) {
+            CursorInRevision findUntil = findBeyondUse ? currentContext->topContext()->range().end : location;
+            declarations = currentContext->findDeclarations(identifier, findUntil);
+
+            for (Declaration* declaration: declarations) {
+                if (declaration->context()->type() != DUContext::Class ||
+                    (currentContext->type() == DUContext::Function && declaration->context() == currentContext->parentContext())) {
+                     // Declarations from class decls must be referenced through `self.<foo>`, except
+                     //  in their local scope (handled above) or when used as default arguments for methods of the same class.
+                     // Otherwise, we're done!
+                    return declaration;
+                }
+            }
+            if (!declarations.isEmpty()) {
+                // If we found declarations but rejected all of them (i.e. didn't return), we need to keep searching.
+                findInNext = true;
+                declarations.clear();
+            }
+        }
+
+        if (!findBeyondUse && currentContext->owner() && currentContext->owner()->isFunctionDeclaration()) {
+            // Names in the body may be defined after the function definition, before the function is called.
+            // Note: only the parameter list has type DUContext::Function, so we have to do this instead.
+            findBeyondUse = findInNext = true;
+        }
+    } while ((currentContext = currentContext->parentContext()));
+
+    return nullptr;
+}
+
+
+Declaration* Helper::declarationForName(const Python::NameAst* name, CursorInRevision location,
+                                        KDevelop::DUChainPointer<const DUContext> context)
+{
+    const Ast* checkNode = name;
+    while ((checkNode = checkNode->parent)) {
+        switch (checkNode->astType) {
+          default:
+            continue;
+          case Ast::ListComprehensionAstType:
+          case Ast::SetComprehensionAstType:
+          case Ast::DictionaryComprehensionAstType:
+          case Ast::GeneratorExpressionAstType:
+            // Variables in comprehensions are used before their definition. `[foo for foo in bar]`
+            auto cmpEnd = CursorInRevision(checkNode->endLine, checkNode->endCol);
+            if (cmpEnd > location) {
+                location = cmpEnd;
+            }
+        }
     }
-    return declaration;
+    return declarationForName(name->identifier->value, location, context);
 }
 
 QVector<DUContext*> Helper::internalContextsForClass(const StructureType::Ptr classType,
@@ -262,9 +288,11 @@ QStringList Helper::getDataDirs() {
     return Helper::dataDirs;
 }
 
-QString Helper::getDocumentationFile() {
-    if ( Helper::documentationFile.isNull() ) {
-        Helper::documentationFile = QStandardPaths::locate(QStandardPaths::GenericDataLocation, "kdevpythonsupport/documentation_files/builtindocumentation.py");
+KDevelop::IndexedString Helper::getDocumentationFile()
+{
+    if ( Helper::documentationFile.isEmpty() ) {
+        auto path = QStandardPaths::locate(QStandardPaths::GenericDataLocation, "kdevpythonsupport/documentation_files/builtindocumentation.py");
+        Helper::documentationFile = IndexedString(path);
     }
     return Helper::documentationFile;
 }
@@ -276,7 +304,7 @@ ReferencedTopDUContext Helper::getDocumentationFileContext()
     }
     else {
         DUChainReadLocker lock;
-        auto file = IndexedString(Helper::getDocumentationFile());
+        auto file = Helper::getDocumentationFile();
         ReferencedTopDUContext ctx = ReferencedTopDUContext(DUChain::self()->chainForDocument(file));
         Helper::documentationFileContext = DUChainPointer<TopDUContext>(ctx.data());
         return ctx;
@@ -450,10 +478,9 @@ QVector<QUrl> Helper::getSearchPaths(const QUrl& workingOnDocument)
         python.start(getPythonExecutablePath(project), getpath);
         python.waitForFinished(1000);
         QString pythonpath = QString::fromUtf8(python.readAllStandardOutput());
-        auto paths = pythonpath.split("$|$");
-        paths.removeAll("");
-        
+
         if ( ! pythonpath.isEmpty() ) {
+            const auto paths = pythonpath.split("$|$", QString::SkipEmptyParts);
             foreach ( const QString& path, paths ) {
                 cachedForProject.append(QUrl::fromLocalFile(path));
             }
